#include "window.h"
#include "aliases/aliases.h"
#include <SDL2/SDL_rect.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_video.h>
#include <stdbool.h>

bool init_window(window_t *wnd, const char *title, u32 width, u32 height, i32 x,
                 i32 y) {
  i32 pos_x = x >= 0 ? x : SDL_WINDOWPOS_CENTERED;
  i32 pos_y = y >= 0 ? y : SDL_WINDOWPOS_CENTERED;

  wnd->window =
      SDL_CreateWindow(title, pos_x, pos_y, width, height, SDL_WINDOW_SHOWN);
  if (!(wnd->window)) {
    return false;
  }

  wnd->renderer = SDL_CreateRenderer(
      wnd->window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
  if (!(wnd->renderer)) {
    cleanup_window(wnd);
    return false;
  }

  wnd->id = SDL_GetWindowID(wnd->window);
  wnd->title = title;

  i32 x_tmp = -1;
  i32 y_tmp = -1;
  SDL_GetWindowPosition(wnd->window, &x_tmp, &y_tmp);

  wnd->x = x_tmp;
  wnd->y = y_tmp;

  i32 w_tmp = -1;
  i32 h_tmp = -1;
  SDL_GetWindowSize(wnd->window, &w_tmp, &h_tmp);

  wnd->width = w_tmp;
  wnd->height = h_tmp;

  return true;
}

void cleanup_window(window_t *wnd) {
  if (wnd->renderer) {
    SDL_DestroyRenderer(wnd->renderer);
    wnd->renderer = NULL;
  }

  if (wnd->window) {
    SDL_DestroyWindow(wnd->window);
    wnd->window = NULL;
  }

  wnd->width = wnd->height = wnd->id = 0;
}

void set_colour(const window_t *wnd, colour_t colour) {
  SDL_SetRenderDrawColor(wnd->renderer, colour.colour.r, colour.colour.g,
                         colour.colour.b, colour.colour.a);
}

void clear_window(const window_t *wnd, colour_t colour) {
  set_colour(wnd, colour);
  SDL_RenderClear(wnd->renderer);
}

void swap_buffers(const window_t *wnd) { SDL_RenderPresent(wnd->renderer); }

void draw_point(const window_t *wnd, point_t p, colour_t colour) {
  set_colour(wnd, colour);
  SDL_RenderDrawPoint(wnd->renderer, p.x, p.y);
}

void draw_line(const window_t *wnd, const line_t *ln, colour_t colour) {
  set_colour(wnd, colour);
  SDL_RenderDrawLine(wnd->renderer, ln->p0.x, ln->p0.y, ln->p1.x, ln->p1.y);
}

void draw_triangle(const window_t *wnd, triangle_t triangle, colour_t colour) {
  line_t ln0 = {triangle.p0, triangle.p1};
  line_t ln1 = {triangle.p0, triangle.p2};
  line_t ln2 = {triangle.p1, triangle.p2};

  draw_line(wnd, &ln0, colour);
  draw_line(wnd, &ln1, colour);
  draw_line(wnd, &ln2, colour);
}

INTERNAL inline i32 min(i32 a, i32 b, i32 c) {
  i32 _min = a <= b ? a : b;
  return _min <= c ? _min : c;
}

INTERNAL inline i32 max(i32 a, i32 b, i32 c) {
  i32 _max = a >= b ? a : b;
  return _max >= c ? _max : c;
}

INTERNAL inline i32 half_space(i32 x1, i32 x2, i32 y1, i32 y2, i32 x, i32 y) {
  return (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1);
}

void fill_triangle(const window_t *wnd, triangle_t triangle, colour_t colour) {
  // Basic triangle filling algorithm from
  // https://web.archive.org/web/20050408192410/http://sw-shader.sourceforge.net/rasterizer.html
  i32 x1 = triangle.p0.x;
  i32 x2 = triangle.p1.x;
  i32 x3 = triangle.p2.x;

  i32 y1 = triangle.p0.y;
  i32 y2 = triangle.p1.y;
  i32 y3 = triangle.p2.y;

  // Find bounding rect
  i32 min_x = min(x1, x2, x3);
  i32 max_x = max(x1, x2, x3);
  i32 min_y = min(y1, y2, y3);
  i32 max_y = max(y1, y2, y3);

  for (i32 y = min_y; y < max_y; ++y) {
    for (i32 x = min_x; x < max_x; ++x) {
      if (half_space(x2, x1, y2, y1, x, y) >= 0 &&
          half_space(x3, x2, y3, y2, x, y) >= 0 &&
          half_space(x1, x3, y1, y3, x, y) >= 0) {
        draw_point(wnd, (point_t){x, y}, colour);
      }
    }
  }
}

void draw_gquad(const window_t *wnd, gquad_t gquad, colour_t colour) {
  line_t l0 = (line_t){gquad.p0, gquad.p1};
  line_t l1 = (line_t){gquad.p1, gquad.p3};
  line_t l2 = (line_t){gquad.p3, gquad.p2};
  line_t l3 = (line_t){gquad.p2, gquad.p0};

  draw_line(wnd, &l0, colour);
  draw_line(wnd, &l1, colour);
  draw_line(wnd, &l2, colour);
  draw_line(wnd, &l3, colour);
}

void fill_gquad(const window_t *wnd, gquad_t gquad, colour_t colour) {
  triangle_t t0 = (triangle_t){gquad.p3, gquad.p1, gquad.p0};
  triangle_t t1 = (triangle_t){gquad.p0, gquad.p2, gquad.p3};

  fill_triangle(wnd, t0, colour);
  fill_triangle(wnd, t1, colour);
}

void draw_rect(const window_t *wnd, rect_t rect, colour_t colour) {
  set_colour(wnd, colour);

  SDL_Rect dst = {rect.topleft.x, rect.topleft.y, rect.w, rect.h};

  SDL_RenderDrawRect(wnd->renderer, &dst);
}

void fill_rect(const window_t *wnd, rect_t rect, colour_t colour) {
  set_colour(wnd, colour);

  SDL_Rect dst = {rect.topleft.x, rect.topleft.y, rect.w, rect.h};

  SDL_RenderFillRect(wnd->renderer, &dst);
}
